import os
import pandas as pd
from dateutil import parser
from IPython import get_ipython
from IPython.display import display, HTML

from lhcsmapi.Time import Time


selected_path = os.path.dirname(__file__)


def _check_feature_comparison(comparison_df):
    if comparison_df['result'].all():
        return True
    else:
        print('The following parameters did not satisfy the comparison criteria:')
        if get_ipython() is None:
            print(comparison_df[comparison_df['result'] == False])
        else:
            display(HTML(comparison_df[comparison_df['result'] == False].to_html()))

        return False


class Reference(object):

    # Query MP3 Quench Excel database columns
    @staticmethod
    def get_mp3_quench_database_columns(circuit_type):
        full_path = os.path.join(selected_path, 'mp3/{}_Quench_Database_Columns.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    # Query table
    @staticmethod
    def get_query_table(circuit_type):
        full_path = os.path.join(selected_path, 'query_table/{}_Query_Table.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    # QH
    @staticmethod
    def get_quench_heater_reference_discharges(circuit_type):
        full_path = os.path.join(selected_path, 'qh/{}_Reference_Discharges.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def get_quench_heater_reference_discharge(circuit_type, cell_name):
        qh_ref_discharges = Reference.get_quench_heater_reference_discharges(circuit_type)
        return qh_ref_discharges[qh_ref_discharges["Cell"] == cell_name]["Timestamp"].values[0]

    @staticmethod
    def find_qh_timestamp_for_magnet_name_from_reference(qh_reference_discharges, magnet):
        # Look for a timestamp of the reference discharge curve
        if qh_reference_discharges[qh_reference_discharges['magnetName'] == magnet].empty:
            return -1
        else:
            return qh_reference_discharges[qh_reference_discharges['magnetName'] == magnet]['timestamp'].values[0]

    @staticmethod
    def get_quench_heaters_reference_features(circuit_type):
        full_path = os.path.join(selected_path, 'qh/{}_Reference_Features.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def get_quench_heaters_reference_differences(circuit_type):
        full_path = os.path.join(selected_path, 'qh/{}_Reference_Differences.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    # EE
    @staticmethod
    def get_energy_extraction_reference_temperature():

        full_path = os.path.join(selected_path, 'ee/RB_Reference_Temperature_Profile.csv')
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def get_energy_extraction_reference_parameters(circuit_type, timestamp_fgc='2021-01-01 00:00:00.000+01:00'):

        if circuit_type == 'RQ':
            full_path = os.path.join(selected_path, 'ee/RQ_Reference_Features.csv')
        elif circuit_type == 'RB':
            date_change = parser.parse('2018-01-01 00:00:00.000+01:00')
            timestamp_fgc_datetime = Time.to_datetime(timestamp_fgc)

            if timestamp_fgc_datetime > date_change:
                full_path = os.path.join(selected_path, 'ee/RB_Reference_Features_2018_01_01.csv')
            else:
                full_path = os.path.join(selected_path, 'ee/RB_Reference_Features_2014_01_01_2018_01_01.csv')
        else:
            raise KeyError('Circuit type {} not supported!'.format(circuit_type))

        return pd.read_csv(full_path, index_col=0)

    # PC
    @staticmethod
    def gety_power_converter_reference_parameters(circuit_type):
        if circuit_type not in ['RB', 'RQ']:
            raise ValueError('Reference features for {} circuit type are not defined.'.format(circuit_type))

        full_path = os.path.join(selected_path, 'pc/{}_Reference_Features.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def get_power_converter_reference_fpa(circuit_type, circuit_name, col):
        if circuit_type == 'RB':
            full_path = os.path.join(selected_path, 'pc/RB_Reference_PNO_b2.csv')
        elif circuit_type == 'RQ':
            full_path = os.path.join(selected_path, 'pc/RQ_Reference_PNO_b3.csv')
        else:
            raise ValueError('Circuit type {} not supported!'.format(circuit_type))

        ref_fpa_df = pd.read_csv(full_path)
        return ref_fpa_df[(ref_fpa_df['circuit'] == circuit_name)][col].values[0]

    # LEADS
    @staticmethod
    def get_current_leads_parameters():
        full_path = os.path.join(selected_path, 'leads/RB_Reference_Threshold.csv')
        return pd.read_csv(full_path, index_col=0)

    # Feature Comparison
    @staticmethod
    def compare_features_to_reference(features_df, circuit_type, system, *, wildcard=None, timestamp=None):
        if system == 'QH':
            reference_df = Reference.get_quench_heaters_reference_features(circuit_type).T
        elif system == 'PC':
            reference_df = Reference.gety_power_converter_reference_parameters(circuit_type).T
        elif system == 'EE':
            reference_df = Reference.get_energy_extraction_reference_parameters(circuit_type, timestamp).T
        else:
            raise KeyError('System {} is not supported'.format(system))

        if wildcard is not None:
            key, value = list(wildcard.items())[0]
            reference_df.columns = [col.replace('%{}%'.format(key), wildcard[key]) for col in reference_df.columns]

        features_df.index = ['act']
        common_columns = list(set(reference_df.columns.values).intersection(set(features_df.columns.values)))
        comparison_df = pd.concat([reference_df[common_columns], features_df[common_columns]], sort=False)
        comparison_df = comparison_df.T
        comparison_df['result'] = comparison_df.apply(
            lambda row: (row['act'] >= row['min']) and (row['act'] <= row['max']),
            axis=1)

        return _check_feature_comparison(comparison_df)


    @staticmethod
    def compare_difference_of_features_to_reference(features_df, reference_df, circuit_type, system, wildcard=None):
        if system == 'QH':
            differences_df = Reference.get_quench_heaters_reference_differences(circuit_type).T
        else:
            raise KeyError('System {} is not supported'.format(system))

        if wildcard is not None:
            key, value = list(wildcard.items())[0]
            differences_df.columns = [col.replace('%{}%'.format(key), value) for col in differences_df.columns]

        features_df.index = ['act']
        reference_df.index = ['ref']

        comparison_df = pd.concat([reference_df[features_df.columns], features_df, differences_df[features_df.columns]],
                                  sort=False)
        comparison_df = comparison_df.T
        comparison_df['result'] = comparison_df.apply(lambda row: abs(row['act']-row['ref']) <= row['diff'], axis=1)

        return _check_feature_comparison(comparison_df)

